from __future__ import print_function, division, absolute_import
from __future__ import unicode_literals
from flask import Flask
import logging
from logging.handlers import RotatingFileHandler
from logging import Formatter
import sys

from orlo.config import config, CONFIG_FILE
from orlo.exceptions import OrloStartupError, OrloError, OrloAuthError, \
    OrloConfigError

try:
    # _version is created by setup.py
    from orlo._version import __version__
except ImportError:
    __version__ = "TEST_BUILD"


app = Flask(__name__)

app.config['SQLALCHEMY_DATABASE_URI'] = config.get('db', 'uri')
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

if config.getboolean('main', 'propagate_exceptions'):
    app.config['PROPAGATE_EXCEPTIONS'] = True

if config.getboolean('db', 'echo_queries'):
    app.config['SQLALCHEMY_ECHO'] = True

if not config.getboolean('main', 'strict_slashes'):
    app.url_map.strict_slashes = False

# Debug mode ignores all custom logging and should only be used in
# local testing...
if config.getboolean('main', 'debug_mode'):
    app.debug = True

if not app.debug:
    log_level = config.get('logging', 'level')
    if log_level == 'debug':
        app.logger.setLevel(logging.DEBUG)
    elif log_level == 'info':
        app.logger.setLevel(logging.INFO)
    elif log_level == 'warning':
        app.logger.setLevel(logging.WARNING)
    elif log_level == 'error':
        app.logger.setLevel(logging.ERROR)

    logfile = config.get('logging', 'file')
    if logfile != 'disabled':
        file_handler = RotatingFileHandler(
            logfile,
            maxBytes=1048576,
            backupCount=1,
        )
        log_format = config.get('logging', 'format')
        formatter = Formatter(log_format)

        file_handler.setFormatter(formatter)
        app.logger.addHandler(file_handler)

if config.getboolean('security', 'enabled') and \
        config.get('security', 'secret_key') == 'change_me':
    raise OrloStartupError(
        "Security is enabled, please configure the secret key")

app.logger.debug("Log level: {}".format(config.get('logging', 'level')))

# Must be imported last
import orlo.error_handlers
import orlo.route_base
import orlo.route_releases
import orlo.route_import
import orlo.route_info
import orlo.route_stats
import orlo.user_auth

app.logger.info("Startup completed with configuration from {}".format(
    CONFIG_FILE))



from __future__ import print_function
from flask import jsonify, request
from orlo import app
from orlo.exceptions import InvalidUsage, OrloError

__author__ = 'alforbes'


@app.errorhandler(404)
def page_not_found(error):
    d = {'message': "404 Not Found", 'url': request.url}
    return jsonify(d), 404


@app.errorhandler(OrloError)
@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@app.errorhandler(400)
def handle_400(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response





from __future__ import print_function
from flask import jsonify, request
from orlo import app
from orlo import __version__

__author__ = 'alforbes'

"""
Base application routes and general functions
"""


@app.before_request
def log_post_data():
    """
    Before each request, log any POST data, without newlines
    """
    s = "{m} {u}".format(m=request.method, u=request.url)
    data = request.get_data(as_text=True)
    if data:
        s += " POST data: {}".format(data.replace('\n', ''))
    app.logger.info(s)


@app.route('/', methods=['GET'])
def root():
    """
    Root page, display info
    """
    return jsonify({
        'message': "Orlo server, see http://orlo.readthedocs.org/"
    })


@app.route('/version', methods=['GET'])
def version():
    """
    Display version information
    :return:
    """
    return jsonify({
        'version': str(__version__)
    })


@app.route('/ping', methods=['GET'])
def ping():
    """
    Simple ping test, takes no parameters

    **Example curl**:

    .. sourcecode:: shell

        curl -X GET 'http://127.0.0.1/ping'
    """
    return jsonify({
        'message': 'pong'
    })



from __future__ import print_function
from flask import request, jsonify, url_for
from orlo import app
import orlo.queries as queries

__author__ = 'alforbes'


def has_no_empty_params(rule):
    defaults = rule.defaults if rule.defaults is not None else ()
    arguments = rule.arguments if rule.arguments is not None else ()
    return len(defaults) >= len(arguments)


@app.route('/info', methods=['GET'])
def info_root():
    """
    Root of /info
    """
    urls = []
    for rule in app.url_map.iter_rules():
        base = str(rule).split('/')[1]
        if base == 'info':
            urls.append(str(rule))

    return jsonify({
        'message': '/info endpoint',
        'urls': sorted(urls),
    }), 200


@app.route('/info/users', methods=['GET'])
@app.route('/info/users/<username>', methods=['GET'])
def info_users(username=None):
    """
    Return a dictionary of users optionally filtering by platform

    :param username: Username to get info for
    :query platform: Platform to filter on
    """
    platform = request.args.get('platform')

    if username:
        users = queries.user_info(username)
    else:
        users = queries.user_summary(platform)

    d = {}
    for user, count in users:
        d[user] = {'releases': count}

    return jsonify(d), 200


@app.route('/info/platforms', methods=['GET'])
@app.route('/info/platforms/<platform>', methods=['GET'])
def info_platforms(platform=None):
    """
    Return a summary of the platforms

    :param platform: Platform to get info for
    """
    if platform:
        platforms = queries.platform_info(platform)
    else:
        platforms = queries.platform_summary()

    d = {}

    for platform, count in platforms:
        d[platform] = {'releases': count}
    return jsonify(d), 200


@app.route('/info/packages', methods=['GET'])
@app.route('/info/packages/<package>', methods=['GET'])
def info_packages(package=None):
    """
    Summary of packages

    :param package: Package to get info for
    :query platform: Platform to filter on
    """
    platform = request.args.get('platform')

    if package:
        packages = queries.package_info(package).all()
    else:
        packages = queries.package_summary(platform=platform).all()

    d = {}

    for package, count in packages:
        d[package] = {'releases': count}

    return jsonify(d), 200


@app.route('/info/packages/list', methods=['GET'])
def info_package_list():
    """
    Return list of all known packages

    :query platform: Platform to filter on
    """

    platform = request.args.get('platform')
    q = queries.package_list(platform=platform)
    result = q.all()
    packages = [r[0] for r in result]
    return jsonify({'packages': packages}), 200


@app.route('/info/packages/versions', methods=['GET'])
def info_package_versions():
    """
    Return current version of all packages

    :query platform:
    """
    platform = request.args.get('platform')
    q = queries.package_versions(platform=platform)
    result = q.all()

    packages = {}
    for package in result:
        packages[package[0]] = package[1]

    return jsonify(packages), 200



from __future__ import print_function
import argparse

try:
    from orlo import __version__
except ImportError:
    # _version.py doesn't exist
    __version__ = "TEST_BUILD"


__author__ = 'alforbes'

"""
Command line interface

Generally setup/initialisation functions and the like, called by /usr/bin/orlo
"""


def parse_args():
    parser = argparse.ArgumentParser(prog='orlo')

    parser.add_argument('--version', '-v', action='version',
                        version='%(prog)s {}'.format(__version__))

    p_config = argparse.ArgumentParser(add_help=False)
    p_config.add_argument('--file', '-f', dest='file_path',
                          help="Config file to read/write",
                          default='/etc/orlo/orlo.ini')

    p_database = argparse.ArgumentParser(add_help=False)
    p_server = argparse.ArgumentParser(add_help=False)
    p_server.add_argument('--host', '-H', dest='host', default='127.0.0.1',
                          help="Address to listen on")
    p_server.add_argument('--port', '-P', dest='port', type=int, default=5000,
                          help="Port to listen on")

    subparsers = parser.add_subparsers(dest='action')
    sp_config = subparsers.add_parser(
            'write_config', help="Write config file",
            parents=[p_config])
    sp_config.set_defaults(func=write_config)

    sp_database = subparsers.add_parser(
            'setup_database', help="Initialise the configured DB",
            parents=[p_database, p_config])
    sp_database.set_defaults(func=setup_database)

    sp_run_server = subparsers.add_parser(
            'run_server', help="Run a test server",
            parents=[p_server, p_config])
    sp_run_server.set_defaults(func=run_server)

    return parser.parse_args()


def write_config(args):
    from orlo import config
    config_file = open(args.file_path, 'w')
    config.write(config_file)


def setup_database(args):
    from orlo.orm import db
    from orlo.config import config

    if config.get('db', 'uri') == 'sqlite://':
        print("Warning: setting up in-memory database, this is "
              "probably not what you want!\n"
              "Please configure db:uri in /etc/orlo/orlo.ini")
    db.create_all()


def run_server(args):
    print("Warning: this is a development server and not suitable "
          "for production, we recommend running under gunicorn.")

    from orlo import app
    app.config['DEBUG'] = True
    app.config['TRAP_HTTP_EXCEPTIONS'] = True
    app.config['PRESERVE_CONTEXT_ON_EXCEPTION'] = False
    app.run(host=args.host, port=args.port, debug=True, use_reloader=True)


def main():
    # Referenced by entry_points in setup.py
    args = parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



from __future__ import print_function
import datetime
import arrow
from orlo import app
from orlo.orm import db, Release, Platform, Package, release_platform
from orlo.exceptions import OrloError, InvalidUsage
from sqlalchemy import and_, exc

__author__ = 'alforbes'

"""
Functions in this file are about generating summaries of data
"""


def get_release(release_id):
    """
    Fetch a single release

    :param release_id:
    :return:
    """
    return db.session.query(Release).filter(Release.id == release_id)


def filter_release_status(query, status):
    """
    Filter the given query by the given release status

    Release status is special, because it's actually determined by the
    package status

    :param query: Query object
    :param status: The status to filter on
    :return:
    """
    enums = Package.status.property.columns[0].type.enums
    if status not in enums:
        raise InvalidUsage("Invalid package status, {} is not in {}".format(
            status, str(enums)))
    if status in ["SUCCESSFUL", "NOT_STARTED"]:
        # ALL packages must match this status for it to apply to the release
        # Query logic translates to "Releases which do not have any packages
        # which satisfy
        # the condition 'Package.status != status'". I.E, all match.
        query = query.filter(
            ~Release.packages.any(
                Package.status != status
            ))
    elif status in ["FAILED", "IN_PROGRESS"]:
        # ANY package can match for this status to apply to the release
        query = query.filter(Release.packages.any(Package.status == status))
    return query


def filter_release_rollback(query, rollback):
    """
    Filter the given query by whether the releases are rollbacks or not

    :param query: Query object
    :param boolean rollback:
    :return:
    """
    if rollback is True:
        # Only count releases which have a rollback package
        query = query.filter(
            Release.packages.any(Package.rollback == True)
        )
    elif rollback is False:
        # Only count releases which do not have any rollback packages
        query = query.filter(
            ~Release.packages.any(Package.rollback == True)
        )
    else:  # What the hell did you pass?
        raise TypeError(
            "Bad rollback parameter: '{}', type {}. Boolean expected.".format(
                rollback, type(rollback)))
    return query


def apply_filters(query, args):
    """
    Apply filters to a query

    :param query: Query object to apply filters to
    :param args: Dictionary of arguments, usually request.args

    :return: filtered query object
    """

    for field, value in args.items():
        if field == 'latest':  # this is not a comparison
            continue

        # special logic for these ones, as they are release attributes that
        # are JIT calculated from package attributes
        if field == 'status':
            query = filter_release_status(query, value)
            continue
        if field == 'rollback':
            query = filter_release_rollback(query, value)
            continue

        if field.startswith('package_'):
            # Package attribute. Ensure source query does a join on Package.
            db_table = Package
            field = '_'.join(field.split('_')[1:])
        else:
            db_table = Release

        comparison = '=='
        time_absolute = False
        time_delta = False
        strip_last = False
        sub_field = None

        if field.endswith('_gt'):
            strip_last = True
            comparison = '>'
        if field.endswith('_lt'):
            strip_last = True
            comparison = '<'
        if field.endswith('_before'):
            strip_last = True
            comparison = '<'
            time_absolute = True
        if field.endswith('_after'):
            strip_last = True
            comparison = '>'
            time_absolute = True
        if 'duration' in field.split('_'):
            time_delta = True
        if field == 'platform':
            field = 'platforms'
            comparison = 'any'
            sub_field = Platform.name

        if strip_last:
            # Strip anything after the last underscore inclusive
            field = '_'.join(field.split('_')[:-1])

        filter_field = getattr(db_table, field)

        # Booleans
        if value in ('True', 'true'):
            value = True
        if value in ('False', 'false'):
            value = False

        # Time related
        if time_delta:
            value = datetime.timedelta(seconds=int(value))
        if time_absolute:
            value = arrow.get(value)

        # Do comparisons
        app.logger.debug(
            "Filtering: {} {} {}".format(filter_field, comparison, value))
        if comparison == '==':
            query = query.filter(filter_field == value)
        if comparison == '<':
            query = query.filter(filter_field < value)
        if comparison == '>':
            query = query.filter(filter_field > value)
        if comparison == 'any':
            query = query.filter(filter_field.any(sub_field == value))

    return query


def releases(**kwargs):
    """
    Return whole releases, based on filters

    :param kwargs: Request arguments
    :return:
    """

    limit = kwargs.pop('limit', None)
    offset = kwargs.pop('offset', None)
    desc = kwargs.pop('desc', False)

    if any(field.startswith('package_') for field in kwargs.keys()) \
            or "status" in kwargs.keys():
        # Package attributes need the join, as does status as it's really a
        # package attribute
        query = db.session.query(Release).join(Package)
    else:
        # No need to join on package if none of our params need it
        query = db.session.query(Release)

    for key in kwargs.keys():
        if isinstance(kwargs[key], bool):
            continue
        if kwargs[key].lower() in ['null', 'none']:
            kwargs[key] = None
    try:
        query = apply_filters(query, kwargs)
    except AttributeError as e:
        raise InvalidUsage(
            "An invalid field was specified: {}".format(e.args[0]))

    if desc:
        stime_field = Release.stime.desc
    else:
        stime_field = Release.stime.asc

    query = query.order_by(stime_field())

    if limit:
        try:
            limit = int(limit)
        except ValueError:
            raise InvalidUsage("limit must be a valid integer value")
        query = query.limit(limit)
    if offset:
        try:
            offset = int(offset)
        except ValueError:
            raise InvalidUsage("offset must be a valid integer value")
        query = query.offset(offset)

    return query


def user_summary(platform=None):
    """
    Find all users that have performed releases and how many

    :param platform: Platform to filter on
    :return: Query result [('user', release_count)]
    """

    query = db.session.query(Release.user, db.func.count(Release.id))
    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))

    return query.group_by(Release.user)


def user_info(username):
    """
    Get user info for a single user

    :param username:
    :return:
    """
    query = db.session.query(
        Release.user, db.func.count(Release.id)) \
        .filter(Release.user == username) \
        .group_by(Release.user)

    return query


def user_list(platform=None):
    """
    Find all users that have performed releases

    :param platform: Platform to filter on
    """
    query = db.session.query(Release.user.distinct())
    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))

    return query


def team_summary(platform=None):
    """
    Find all teams that have performed releases and how many

    :param platform: Platform to filter on
    :return: Query result [('team', release_count)]
    """

    query = db.session.query(Release.team, db.func.count(Release.id))
    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))

    return query.group_by(Release.team)


def team_info(team_name):
    """
    Get info for a single team

    :param team_name:
    :return:
    """
    query = db.session.query(
        Release.user, db.func.count(Release.id)) \
        .filter(Release.team == team_name) \
        .group_by(Release.team)

    return query


def team_list(platform=None):
    """
    Find all teams that have performed releases

    :param platform: Platform to filter on
    """
    query = db.session.query(Release.team.distinct())
    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))

    return query


def package_summary(platform=None, stime=None, ftime=None):
    """
    Summary of packages

    :param stime: Start time, or time lower bound
    :param ftime: Finish time, or time upper bound
    :param platform: Filter by platform
    """

    query = db.session.query(Package.name, db.func.count(Package.release_id))
    if any(x is not None for x in [platform, stime, ftime]):
        query = query.join(Release)
    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))
    if stime:
        query = query.filter(Release.stime > stime)
    if ftime:
        query = query.filter(Release.stime < ftime)

    query = query.group_by(Package.name)

    return query


def package_info(package_name):
    """
    Return a query for a package and how many times it was released

    :param package_name:
    :return:
    """
    query = db.session.query(
        Package.name, db.func.count(Package.id)) \
        .filter(Package.name == package_name) \
        .group_by(Package.name)

    return query


def package_list(platform=None):
    """
    Find all packages that have been released

    :param platform: Platform to filter on
    """
    query = db.session.query(Package.name).distinct()
    if platform:
        query = query.join(Release).filter(
            Release.platforms.any(Platform.name == platform))

    return query


def package_versions(platform=None):
    """
    List the current version of all packages

    It is not sufficient to just return the highest version of each successful
    package, as they can be rolled back, so we determine the version by last
    release time

    :param platform: Platform to filter on
    """

    # Sub query gets a list of successful packages by last successful release
    # time
    sub_q = db.session.query(
            Package.name.label('name'),
            db.func.max(Package.stime).label('max_stime')) \
        .filter(Package.status == 'SUCCESSFUL')
    if platform:  # filter by platform
        sub_q = sub_q \
            .join(Release) \
            .filter(Release.platforms.any(Platform.name == platform))
    sub_q = sub_q \
        .group_by(Package.name) \
        .subquery()

    # q inner-joins Package table with sub_q to get the version
    q = db.session.query(
            Package.name,
            Package.version) \
        .join(sub_q, and_(sub_q.c.max_stime == Package.stime,
                          sub_q.c.name == Package.name)) \
        .group_by(Package.name, Package.version)

    return q


def count_releases(user=None, package=None, team=None, platform=None,
                   status=None, rollback=None, stime=None, ftime=None):
    """
    Return the number of releases with the attributes specified

    :param string user: Filter by username
    :param string package:  Filter by package
    :param string team:  Filter by team
    :param string status:  Filter by status
    :param string platform: Filter by platform
    :param string stime: Filter by releases that started after
    :param string ftime: Filter by releases that started before
    :param boolean rollback: Filter on whether or not the release contains a \
    rollback
    :return: Query

    Note that rollback and status are special fields when applied to a
    release, as they are Package attributes.

    A "successful" or "in progress" release is defined as a release where all
    packages match the status. Conversely, a "failed" or "not started" release
    is defined as a release where any package matches.

    For rollbacks, if any package is a rollback the release is included,
    otherwise if all packages are not rollbacks the release obviously isn't
    either.

    Implication of this is that a release can be both "failed" and "in
    progress".
    """

    args = {
        'user': user, 'package': package, 'team': team, 'platform': platform,
        'status': status, 'rollback': rollback, 'stime': stime, 'ftime': ftime,
    }
    app.logger.debug("Entered count_releases with args: {}".format(str(args)))

    query = db.session.query(db.func.count(Release.id.distinct())).join(Package)

    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))
    if user:
        query = query.filter(Release.user == user)
    if team:
        query = query.filter(Release.team == team)
    if package:
        query = query.filter(Package.name == package)
    if stime:
        query = query.filter(Release.stime >= stime)
    if ftime:
        query = query.filter(Release.stime <= ftime)

    if rollback is not None:
        query = filter_release_rollback(query, rollback)

    if status:
        query = filter_release_status(query, status)

    return query


def count_packages(user=None, team=None, platform=None, status=None,
                   rollback=None):
    """
    Return the number of packages with the attributes specified

    :param string user:
    :param string team:
    :param string platform:
    :param string status:
    :param boolean rollback:
    :return: Query
    """

    query = db.session.query(db.func.count(Package.id.distinct())).join(Release)

    if platform:
        query = query.filter(Release.platforms.any(Platform.name == platform))
    if user:
        query = query.filter(Release.user == user)
    if team:
        query = query.filter(Release.team == team)
    if status:
        query = query.filter(Package.status == status)
    if rollback is not None:
        query = query.filter(Package.rollback == rollback)

    return query


def platform_summary():
    """
    Return a summary of the known platforms

    :return: Query result [('platform'), release_count]
    """

    query = db.session.query(
        Platform.name, db.func.count(Platform.id)) \
        .join(release_platform) \
        .group_by(Platform.name)

    return query


def platform_info(platform_name):
    """
    Return a single platform and how many times it was released

    :param platform_name:
    :return:
    """

    query = db.session.query(
        Platform.name, db.func.count(Platform.id)) \
        .filter(Platform.name == platform_name) \
        .group_by(Platform.name)

    return query


def platform_list():
    """
    Return a summary of the known platforms

    :return: Query result [('platform'), release_count]
    """

    query = db.session.query(Platform.name.distinct()) \
        .join(release_platform)

    return query



from __future__ import print_function
from orlo.queries import apply_filters, filter_release_rollback, filter_release_status
from orlo import app
from orlo.orm import db, Release, Platform, Package, release_platform
from orlo.exceptions import OrloError, InvalidUsage
from collections import OrderedDict

__author__ = 'alforbes'

"""
Functions related to building statistics
"""


def releases_by_time(unit, summarize_by_unit=False, **kwargs):
    """
    Return stats by time from the given arguments

    :param summarize_by_unit: Passed to add_release_by_time_to_dict()
    :param unit: Passed to add_release_by_time_to_dict()
    """

    query = db.session.query(Release.id, Release.stime).join(Package).group_by(Release)
    query = apply_filters(query, kwargs)

    return get_dict_of_objects_by_time(query, unit, summarize_by_unit)


def packages_by_time(unit, summarize_by_unit=False, **kwargs):
    """
    Count packages by time from the filters given

    :param summarize_by_unit: Passed to add_release_by_time_to_dict()
    :param unit: Passed to add_release_by_time_to_dict()
    """

    query = db.session.query(Package.id, Package.name, Package.stime).join(Release)
    query = apply_filters(query, kwargs)

    return get_dict_of_objects_by_time(query, unit, summarize_by_unit)


# TODO add stats_user_time and stats_team_time
# or generalise a stats_time function


def get_dict_of_objects_by_time(query, unit, summarize_by_unit=False):
    """
    Build a dictionary which summarises the objects in the query given

    :param query:
    :param unit:
    :param summarize_by_unit:
    :return:
    """

    # Build queries for the individual stats
    q_normal_successful = filter_release_status(
            filter_release_rollback(query, rollback=False), 'SUCCESSFUL'
    )
    q_normal_failed = filter_release_status(
            filter_release_rollback(query, rollback=False), 'FAILED'
    )
    q_rollback_successful = filter_release_status(
            filter_release_rollback(query, rollback=True), 'SUCCESSFUL'
    )
    q_rollback_failed = filter_release_status(
            filter_release_rollback(query, rollback=True), 'FAILED'
    )

    output_dict = OrderedDict()

    add_objects_by_time_to_dict(
            q_normal_successful, output_dict, ('normal', 'successful'), unit, summarize_by_unit)
    add_objects_by_time_to_dict(
            q_normal_failed, output_dict, ('normal', 'failed'), unit, summarize_by_unit)
    add_objects_by_time_to_dict(
            q_rollback_successful, output_dict, ('rollback', 'successful'), unit,
            summarize_by_unit)
    add_objects_by_time_to_dict(
            q_rollback_failed, output_dict, ('rollback', 'failed'), unit, summarize_by_unit)

    return output_dict


def add_objects_by_time_to_dict(query, releases_dict, t_category, unit='month',
                                summarize_by_unit=False):
    """
    Take a query and add each of its objects to a dictionary, broken down by time

    If the query given has a 'name' column, that will be included in the dictionary path
    above the categories (t_category).

    :param dict releases_dict: Dict to add to
    :param tuple t_category: tuple of headings, i.e. (<normal|rollback>, <successful|failed>)
    :param query query: Query object to retrieve releases from
    :param string unit: Can be 'iso', 'hour', 'day', 'week', 'month', 'year',
    :param boolean summarize_by_unit: Only break down releases by the given unit, i.e. only one
        layer deep. For example, if "year" is the unit, we group all releases under the year
        and do not add month etc underneath.
    :return:

    **Note**: this can also be use for packages
    """
    app.logger.debug("Entered add_objects_by_time_to_dict")
    for object_ in query:
        if summarize_by_unit:
            tree_args = [str(getattr(object_.stime, unit))]
        else:
            if unit == 'year':
                tree_args = [str(object_.stime.year)]
            elif unit == 'month':
                tree_args = [str(object_.stime.year), str(object_.stime.month)]
            elif unit == 'week':
                # First two args of isocalendar(), year and week
                tree_args = [str(i) for i in object_.stime.isocalendar()][0:2]
            elif unit == 'iso':
                tree_args = [str(i) for i in object_.stime.isocalendar()]
            elif unit == 'day':
                tree_args = [str(object_.stime.year), str(object_.stime.month),
                             str(object_.stime.day)]
            elif unit == 'hour':
                tree_args = [str(object_.stime.year), str(object_.stime.month),
                             str(object_.stime.day), str(object_.stime.hour)]
            else:
                raise InvalidUsage(
                        'Invalid unit "{}" specified for release breakdown'.format(unit))
        if hasattr(object_, 'name'):
            #
            tree_args.append(object_.name)
        # Append categories
        tree_args += t_category
        append_tree_recursive(releases_dict, tree_args[0], tree_args)


def append_tree_recursive(tree, parent, nodes, node_index=0):
    """
    Recursively place the nodes under each other

    :param dict tree: The dictionary we are operating on
    :param parent: The parent for this node
    :param nodes: The list of nodes
    :param node_index: The position in the list list we are up to
    :return:
    """
    app.logger.debug('Called recursive function with args:\n{}, {}, {}'.format(
            str(tree), str(parent), str(nodes)))

    child_index = node_index + 1
    try:
        # Get the child, one after the parent
        child = nodes[child_index]
    except IndexError:
        # Must be at end
        if parent in tree:
            tree[parent] += 1
        else:
            tree[parent] = 1
        return tree

    # Otherwise recurse again
    if parent not in tree:
        tree[parent] = {}
    # Child becomes the parent
    append_tree_recursive(tree[parent], child, nodes, node_index=child_index)



from __future__ import print_function, unicode_literals
from flask import json
from orlo import app
from orlo.orm import db, Release, Package, Platform
from orlo.exceptions import InvalidUsage
from sqlalchemy.orm import exc
from six import string_types
import uuid

__author__ = 'alforbes'


def append_or_create_platforms(request_platforms):
    """
    Create the platforms if they don't exist, and return a list of Platform
    objects

    :param list request_platforms: List of strings denoting platform names
    """
    platforms = []
    for p in request_platforms:
        try:
            query = db.session.query(Platform).filter(Platform.name == p)
            platform = query.one()
            # app.logger.debug("Found platform {}".format(platform.name))
        except exc.NoResultFound:
            app.logger.info("Creating platform {}".format(p))
            platform = Platform(p)
            db.session.add(platform)
        platforms.append(platform)
    return platforms


def create_release(request):
    """
    Create a Release object from a request
    """

    references = request.json.get('references', [])
    if references and type(references) is list:
        references = list_to_string(references)

    # If given a single string, make it a list
    request_platforms = request.json.get('platforms')
    if request_platforms and type(request_platforms) is not list:
        request_platforms = [request_platforms]

    # Get the platforms
    platforms = append_or_create_platforms(request_platforms)

    release = Release(
        # Required attributes
        platforms=platforms,
        user=request.json['user'],
        # Not required
        team=request.json.get('team'),
        references=references,
    )
    return release


def create_package(release_id, request):
    """
    Create a package object for a release
    """

    return Package(
        release_id,
        request.json.get('name'),
        request.json.get('version'),
        diff_url=request.json.get('diff_url', None),
        rollback=request.json.get('rollback', False),
    )


def fetch_release(release_id):
    """
    Fetch a release by ID
    """
    rq = db.session.query(Release).filter(Release.id == release_id)
    release = rq.first()
    if not release:
        raise InvalidUsage("Release does not exist")
    return release


def fetch_package(release_id, package_id):
    """
    Fetch a package, and validate it is part of the release
    """

    # Fetch release first, as it is a good sanity check
    fetch_release(release_id)
    pq = db.session.query(Package).filter(Package.id == package_id)
    package = pq.first()

    if not package:
        raise InvalidUsage("Package does not exist")
    if str(package.release_id) != release_id:
        raise InvalidUsage("This package does not belong to this release")

    return package


def validate_request_json(request):
    try:
        request.json
    except Exception:
        # This is pretty ugly, but we want something more user friendly than
        # "Bad Request"
        raise InvalidUsage("Could not parse JSON document")

    if not request.json:
        raise InvalidUsage('Missing application/json header', status_code=400)


def validate_release_input(request):
    validate_request_json(request)
    if 'platforms' not in request.json:
        raise InvalidUsage('JSON doc missing platforms field', status_code=400)
    return True


def validate_package_input(request, release_id):
    validate_request_json(request)

    if not 'name' in request.json or not 'version' in request.json:
        raise InvalidUsage("Missing name / version in request body.")
    app.logger.debug(
        "Package request validated, release_id {}".format(release_id))
    return True


def _validate_package_stop_input(request):
    validate_request_json(request)
    if 'success' not in request.json:
        raise InvalidUsage("Missing success key in JSON doc")


def list_to_string(array):
    """
    Convert a list to a string for storage in the DB

    :param array:
    :return:
    """
    if not array:
        return []

    return '["' + '", "'.join(array) + '"]'


def stream_json_list(heading, iterator):
    """
    A lagging generator to stream JSON so we don't have to hold everything in
    memory

    This is a little tricky, as we need to omit the last comma to make valid
    JSON, thus we use a lagging generator, similar to
    http://stackoverflow.com/questions/1630320/

    :param heading: The title of the set, e.g. "releases"
    :param iterator: Any object with __iter__(), e.g. SQLAlchemy Query
    """
    iterator = iterator.__iter__()
    try:
        prev_release = next(iterator)  # get first result
    except StopIteration:
        # StopIteration here means the length was zero, so yield a valid
        # releases doc and stop
        yield '{{"{}": []}}'.format(heading)
        raise StopIteration

    # We have some releases. First, yield the opening json
    yield '{{"{}": ['.format(heading)

    # Iterate over the releases
    for item in iterator:
        yield json.dumps(prev_release.to_dict()) + ', '
        prev_release = item

    # Now yield the last iteration without comma but with the closing brackets
    yield json.dumps(prev_release.to_dict()) + ']}'


def str_to_bool(value):
    if isinstance(value, string_types):
        try:
            value = int(value)
        except ValueError:
            if value.lower() in ('t', 'true'):
                return True
            elif value.lower() in ('f', 'false'):
                return False
    if isinstance(value, int):
        return True if value > 0 else False
    raise ValueError("Value {} can not be cast as boolean".format(value))


def is_uuid(string):
    """
    Test whether a string is a valid UUID

    :param string:
    :return:
    """
    try:
        uuid.UUID(string)
    except ValueError:
        return False
    return True



from flask_sqlalchemy import SQLAlchemy
from sqlalchemy_utils.types.uuid import UUIDType
from sqlalchemy_utils.types.arrow import ArrowType

from orlo import app, config
from orlo.exceptions import OrloWorkflowError
import pytz
import uuid
import arrow
import json

__author__ = 'alforbes'

db = SQLAlchemy(app)

try:
    TIMEZONE = pytz.timezone(config.get('main', 'time_zone'))
except pytz.exceptions.UnknownTimeZoneError:
    app.logger.critical(
        'Unknown time zone "{}", see pytz docs for valid values'.format(
            config.get('main', 'timezone')
        ))


# Map releases to platforms
release_platform = db.Table(
    'release_platform', db.Model.metadata,
    db.Column('release_id', UUIDType, db.ForeignKey('release.id')),
    db.Column('platform_id', UUIDType, db.ForeignKey('platform.id'))
)


def string_to_list(string):
    """
    Load a list from a string

    :param string:
    :return:
    """
    if string is None:
        return []

    if '[' in string and ']' in string and ('"' in string or "'" in string):
        # Valid list syntax, presumably...
        # TODO a regex would be better here
        return json.loads(string.replace("'", '"'))
    else:
        # assume just one item
        return [string]


class Release(db.Model):
    """
    The main Release object
    """
    __tablename__ = 'release'

    id = db.Column(UUIDType, primary_key=True, unique=True, nullable=False)
    platforms = db.relationship('Platform', secondary=release_platform)
    references = db.Column(db.String)
    stime = db.Column(ArrowType, index=True)
    ftime = db.Column(ArrowType)
    duration = db.Column(db.Interval)
    user = db.Column(db.String, nullable=False)
    team = db.Column(db.String)
    packages = db.relationship("Package", backref=db.backref("release"))

    def __init__(self, platforms, user, team=None, references=None):
        # platforms and references are stored as strings in DB but really are lists
        self.id = uuid.uuid4()
        self.platforms = platforms
        self.user = user

        if team:
            self.team = team
        if references:
            self.references = str(references)

        # Assume the release started when it was created
        self.start()

    def __str__(self):
        return self.to_dict()

    def to_dict(self):
        time_format = config.get('main', 'time_format')

        metadata = {}
        for m in self.metadata:
            metadata.update(m.to_dict())

        return {
            'id': str(self.id),
            'packages': [p.to_dict() for p in self.packages],
            'platforms': [platform.name for platform in self.platforms],
            'references': string_to_list(self.references),
            'stime': self.stime.strftime(config.get('main', 'time_format')) if self.stime else None,
            'ftime': self.ftime.strftime(config.get('main', 'time_format')) if self.ftime else None,
            'duration': self.duration.seconds if self.duration else None,
            'metadata': metadata,
            'user': self.user,
            'team': self.team,
        }

    def start(self):
        """
        Mark a release as started
        """
        self.stime = arrow.now(config.get('main', 'time_zone'))

    def stop(self):
        """
        Mark a release as stopped
        """
        self.ftime = arrow.now(config.get('main', 'time_zone'))
        td = self.ftime - self.stime
        self.duration = td


class Package(db.Model):
    """
    A deployed instance of a package
    """
    __tablename__ = 'package'

    id = db.Column(UUIDType, primary_key=True, unique=True, nullable=False)
    name = db.Column(db.String(120), nullable=False)
    stime = db.Column(ArrowType, index=True)
    ftime = db.Column(ArrowType)
    duration = db.Column(db.Interval)
    status = db.Column(
        db.Enum('NOT_STARTED', 'IN_PROGRESS', 'SUCCESSFUL', 'FAILED',
                name='status_types'),
        default='NOT_STARTED')
    version = db.Column(db.String(32), nullable=False)
    diff_url = db.Column(db.String)
    rollback = db.Column(db.Boolean(create_constraint=True))
    release_id = db.Column(UUIDType, db.ForeignKey("release.id"))

    def __init__(self, release_id, name, version, diff_url=None, rollback=False):
        self.id = uuid.uuid4()
        self.name = name
        self.version = version
        self.release_id = release_id
        self.diff_url = diff_url
        self.rollback = rollback

    def start(self):
        """
        Mark a package deployment as started
        """
        self.stime = arrow.now(config.get('main', 'time_zone'))
        self.status = 'IN_PROGRESS'

    def stop(self, success):
        """
        Mark a package deployment as stopped

        :param success: Whether or not the package deploy succeeded
        """
        if self.stime is None:
            raise OrloWorkflowError(
                "Can not stop a package which has not been started")
        self.ftime = arrow.now(config.get('main', 'time_zone'))

        td = self.ftime - self.stime
        self.duration = td

        if success:
            self.status = 'SUCCESSFUL'
        else:
            self.status = 'FAILED'

    def to_dict(self):
        time_format = config.get('main', 'time_format')
        return {
            'id': str(self.id),
            'name': self.name,
            'version': self.version,
            'stime': self.stime.strftime(config.get('main', 'time_format')) if self.stime else None,
            'ftime': self.ftime.strftime(config.get('main', 'time_format')) if self.ftime else None,
            'duration': self.duration.seconds if self.duration else None,
            'rollback': self.rollback,
            'status': self.status,
            'diff_url': self.diff_url,
        }


class PackageResult(db.Model):
    """
    The results of a package
    """
    __tablename__ = 'package_result'

    id = db.Column(UUIDType, primary_key=True, unique=True)
    content = db.Column(db.Text)

    package_id = db.Column(UUIDType, db.ForeignKey("package.id"))
    package = db.relationship("Package", backref=db.backref('results',
                                                            order_by=id))

    def __init__(self, package_id, content):
        self.id = uuid.uuid4()
        self.package_id = package_id
        self.content = content


class ReleaseNote(db.Model):
    """
    Notes added to a release
    """
    __tablename__ = 'release_note'

    id = db.Column(UUIDType, primary_key=True, unique=True)
    content = db.Column(db.Text, nullable=False)

    release_id = db.Column(UUIDType, db.ForeignKey("release.id"))
    release = db.relationship("Release", backref=db.backref('notes', order_by=id))

    def __init__(self, release_id, content):
        self.id = uuid.uuid4()
        self.release_id = release_id
        self.content = content


class ReleaseMetadata(db.Model):
    """
    Metadata added to a release
    """
    __tablename__ = 'release_metadata'

    id = db.Column(UUIDType, primary_key=True, unique=True)

    release_id = db.Column(UUIDType, db.ForeignKey("release.id"))
    release = db.relationship("Release", backref=db.backref('metadata', order_by=id))
    key = db.Column(db.Text, nullable=False)
    value = db.Column(db.Text, nullable=False)

    def __init__(self, release_id, key, value):
        self.id = uuid.uuid4()
        self.release_id = release_id
        self.key = key
        self.value = value

    def getKey(self):
        return str(self.key)

    def to_dict(self):
        return {self.key: self.value}


class Platform(db.Model):
    """
    A platform that is released to
    """
    __tablename__ = 'platform'

    id = db.Column(UUIDType, primary_key=True, unique=True)
    name = db.Column(db.Text, nullable=False, unique=True)

    def __init__(self, name):
        self.id = uuid.uuid4()
        self.name = name



import arrow
import json
from flask import jsonify, request
from orlo import app
from orlo.orm import db, Package, Release, PackageResult, ReleaseNote, Platform
from orlo.util import validate_request_json
from orlo.user_auth import token_auth
from sqlalchemy.orm import exc

__author__ = 'alforbes'


@token_auth.token_required
@app.route('/releases/import', methods=['GET'])
def get_import():
    """
    Display a useful message on how to import

    :status 200:
    """

    msg = 'This is the orlo import url'
    return jsonify(message=msg), 200


@app.route('/releases/import', methods=['POST'])
def post_import():
    """
    Import a release.

    This endpoint is designed to created releases in bulk. It bypasses the normal workflow,
    and may be an option for those who wish to "publish" a release after the fact rather than step
    through as it happens.

    The document must contain a list of full releases, in json format. Example:

    .. code-block:: python

        [
            {
                "platforms": [
                  "GumtreeUK"
                ],
                "stime": "2015-12-17T17:02:04Z",
                "ftime": "2015-12-17T17:02:24Z",
                "team": "Gumtree UK Site Operations",
                "references": [
                  "TICKET-1"
                ],
                "notes": [
                  "Imported from other_system"
                ],
                "packages": [
                    {
                        "name": "",
                        "diff_url": null,
                        "stime": "2015-12-17T17:02:22Z"
                        "ftime": 1450371742,
                        "rollback": false,
                        "status": "SUCCESSFUL",
                        "version": "1.0.1",
                    }
                ],
                "user": "user_one"
            },
            {...}
        ]

    Timestamps can be any format understood by Arrow (note mix of unix time and ISO timestamps
    above).

    Status must be one of the enums accepted by `orlo.orm.Package.status`, i.e.:
    NOT_STARTED, IN_PROGRESS, SUCCESSFUL, FAILED

    A json null value is acceptable for non-required fields, or it can be omitted entirely.
    See `orlo.orm.Release` and `orlo.orm.Package`.

    **Example curl**:

    .. sourcecode:: shell

        curl -v -X POST -d @releases.json 'http://127.0.0.1:5000/releases/import' -H \
        "Content-Type: application/json"

    :status 200: The document was accepted
    """

    validate_request_json(request)

    releases = []
    for r in request.json:
        # Get the platform, create if it doesn't exist
        platforms = []
        for p in r['platforms']:
            try:
                query = db.session.query(Platform).filter(Platform.name == p)
                platform = query.one()
            except exc.NoResultFound:
                app.logger.info("Creating platform {}".format(p))
                platform = Platform(p)
                db.session.add(platform)
            platforms.append(platform)

        release = Release(
                platforms=platforms,
                user=r['user'],
                team=r.get('team'),
                references=json.dumps(r.get('references')),
        )

        release.stime = arrow.get(r['stime']) if r.get('stime') else None
        release.ftime = arrow.get(r['ftime']) if r.get('ftime') else None
        if release.ftime and release.stime:
            release.duration = release.ftime - release.stime

        notes = r.get('notes')
        if notes:
            for n in notes:
                note = ReleaseNote(release.id, n)
                db.session.add(note)

        for p in r['packages']:
            package = Package(
                    release_id=release.id,
                    name=p['name'],
                    version=p['version'],
            )

            package.rollback = p.get('rollback')
            package.status = p.get('status')
            package.diff_url = p.get('diff_url')

            if p.get('stime'):
                package.stime = arrow.get(p['stime'])
            else:
                package.stime = arrow.get(r['stime'])
            package.ftime = arrow.get(p['ftime']) if p.get('ftime') else None
            if package.stime and package.ftime:
                package.duration = package.ftime - package.stime

            db.session.add(package)

        db.session.add(release)
        db.session.commit()

        releases.append(release.id)

    return jsonify({'releases': [str(x) for x in releases]}), 200
